/*******************************************************************************
BoundingDiameters
@author Frank Takes - ftakes@liacs.nl
@website http://www.liacs.nl/~ftakes/diameter
**************** consult the README.txt file before using *********************/

////////////////////////////////////////////////////////////////////////////////
// sorting functions

bool byupperordegree(int a, int b) {
	if(ecc_upper[a] == ecc_upper[b]) 
		return OUTDEG[a] < OUTDEG[b];
	return ecc_upper[a] < ecc_upper[b];	
}

bool bylowerordegree(int a, int b) {
	if(ecc_lower[a] == ecc_lower[b])
		return OUTDEG[a] > OUTDEG[b];
	return ecc_lower[a] < ecc_lower[b];
}


bool bybounddiff(int a, int b) {
	if(abs(ecc_upper[a]-ecc_lower[a]) == abs(ecc_upper[b]-ecc_lower[b])) 
		return bylowerordegree(a, b);
	return abs(ecc_upper[a]-ecc_lower[a]) < abs(ecc_upper[b]-ecc_lower[b]);
}

bool bydistanceordegree(int a, int b) {
	if(distances[a] == distances[b]) 
		return OUTDEG[a] < OUTDEG[b];
	return distances[a] < distances[b];	
}

bool bydegree(int a, int b) {
	return OUTDEG[a] < OUTDEG[b];
}

////////////////////////////////////////////////////////////////////////////////

void Graph::preprocess() {  
	int sum, marked = 0;
	for(int i=1; i<=N; i++)
		if(exists(i) && WCC[i] == largestWCC) {
			sum = 0;
			for(int j=0; j<(signed)E[i].size(); j++) 
				if(OUTDEG[E[i][j]] == 1) { // prune identical subtrees of size 1
					sum++; 
					if(sum > 1) {
						visited[E[i][j]] = true;
						marked++;
					}
				} // if
		} // if
	outputStat("Nodes pruned in preprocessing", marked);
} // preprocess


void Graph::selectFrom(int & a, int selectionMethod) {
	static bool takelow = false;
	
	// first iteration: take highest degree
	if(a == -1) {
		a = 1;
		// find largest
		for(int i=1; i<=N; i++) {
			if(OUTDEG[i] > OUTDEG[a])
				a = i;
			ecc_upper[i] = 10000;
			ecc_lower[i] = -1;		
			visited[i] = false;
		} // for			
	}
	
	// prepare list of candidates (could also directly run through this list, 
	// using O(n) time, but this is more flexible (and only O(n log(n)))
	vector<int> tochoose;
	tochoose.clear();
	for(int i=1; i<=N; i++) 
		if(!visited[i] && ((WCC[i] == largestWCC)))
			tochoose.push_back(i);		
	
	switch(selectionMethod) { // based on function parameter, select next node
		case 0:
			a = tochoose[rand()%((signed)tochoose.size())];
		break;
		case 1: // repeatedly take least explored area
			sort(tochoose.begin(), tochoose.end(), bybounddiff);			
			for(int j=(signed)tochoose.size()-1; j>=0; j--) {
				a = tochoose[j];
				break;
			} // if
		break;			
		case 2: // repeatedly take lowest- or highest lower- or upper bound
			if(!takelow) {
				sort(tochoose.begin(), tochoose.end(), byupperordegree);		
				for(int j=(signed)tochoose.size()-1; j>=0; j--) {
					a = tochoose[j];
					break;
				} // if
			} // if
			else {
				sort(tochoose.begin(), tochoose.end(), bylowerordegree);				
				for(int j=0; j<(signed)tochoose.size(); j++) {
					a = tochoose[j];
					break;
				} // if
			} // else
			takelow = !takelow;	
		break;
		case 5: // volgende node selecteren: steeds verste nemen
			sort(tochoose.begin(), tochoose.end(), bydistanceordegree);	//		
			for(int j=(signed)tochoose.size()-1; j>=0; j--) {
				a = tochoose[j];
				break;
			} // if
			
		break;	
		case 4: // repeatedly take lowest- or highest degree
			if(!takelow) {
				sort(tochoose.begin(), tochoose.end(), bydegree);	// how to sort?			
				for(int j=(signed)tochoose.size()-1; j>=0; j--) {
					a = tochoose[j];
					break;
				} // if
			} // if
			else {
				sort(tochoose.begin(), tochoose.end(), bydegree);	// how to sort?					
				for(int j=0; j<(signed)tochoose.size(); j++) {
					a = tochoose[j];
					break;
				} // if
			} // else
			takelow = !takelow;	
		break;				
		default: a = -1; break;
	} // switch
	
} // selectFrom

int Graph::eccentricity(int a, int * distances) {
	int temp, current, ecc_current=0;
	

	// reset arrays
	for(int i=1; i<=N; i++) {
		distances[i] = N;
		previous[i] = -1;			
	}
	for(int i=0; i<N; i++) {
		dists[i].clear();
	} // for	
	
	queue<int> qu;
	while(!qu.empty())
		qu.pop();
	distances[a] = 0;
	qu.push(a);

	while(!qu.empty()) {
		current = qu.front();
		qu.pop();
		for(int i=0; i<(signed)E[current].size(); i++) {
			if(distances[E[current][i]] == N) {
				temp = E[current][i];
				distances[temp] = distances[current] + 1;
				previous[temp] = current;
				dists[distances[current] + 1].push_back(temp);
				qu.push(temp);
			}
		} // for
	} // whiles
	
	// get current eccentricity
	for(int i=0; i<MAXN; i++)
		if(dists[i].size() > 0)
			if(i > ecc_current) 
				ecc_current = i;
	return ecc_current;
} // longestSSP;


int Graph::diameter(int selectionMethod = 2, bool directed=false) {

	print(); 
	preprocess();
	cerr << endl;
	
	int lower = -N, upper = N, a = -1, reallyknown=-1, 
		unvisited=N, numvis=0, ecc_current, maxupper;
	vector<int> solpath;
	int longest = -1;	

	cerr <<  "   i low/upp      unvisited             ecc. known          next node" << endl;
	
	while(unvisited > 0 && lower < upper) {
		selectFrom(a, selectionMethod);
		visited[a] = true;
		numvis++;
		
		ecc_current = eccentricity(a, distances);
		
		// give summary
		cerr << setw(4) << numvis << " "  
		<< setw(4) << lower << "/"
		<< setw(2) << upper 
	    << setw(10) << unvisited << " (" 
	    << setw(10) << setprecision(3) << ((double)unvisited*100.0)/(double)realN << "%) "
	    << setw(7) << reallyknown << " (" 
	    << setw(10) << setprecision(3)  << ((double)reallyknown*100.0)/(double)realN << "%)  #" 
		<< setw(7) << a  
		<< " (deg: "  << setw(6) << OUTDEG[a] << ", bounds: " << setw(2) << ecc_lower[a] << "/" 
		<< setw(2) << ecc_upper[a] << "): " << setw(3) << ecc_current << endl;			
		// end of summary 
				

		// save the path
		if(ecc_current > longest) {
			longest = ecc_current;
			solpath.clear();
			int c = dists[longest][0];
			while(previous[c] != -1) {
				solpath.push_back(c);
				c = previous[c];
			}
			solpath.push_back(c);			
		} // if it is longest path so far
		
		if(ecc_current > lower)
			lower = ecc_current;
		if(ecc_current*2 < upper) 
			upper = ecc_current*2;
		
		reallyknown = 0;
		unvisited = 0;
		maxupper = 0;

		for(int i=1; i<=N; i++) {
			if(exists(i) && WCC[i] == largestWCC) {

				ecc_lower[i] = max(distances[i], max(ecc_current-distances[i], ecc_lower[i]));	
				ecc_upper[i] = min(ecc_current+distances[i], ecc_upper[i]);
				lower = max(lower, ecc_lower[i]);
				maxupper = max(maxupper, ecc_upper[i]);

				if(ecc_lower[i] == ecc_upper[i] || ((ecc_upper[i] <= lower) && ((double)ecc_lower[i] >= (double)upper/2.0))) {
					if(ecc_lower[i] == ecc_upper[i])
						reallyknown++;		
					visited[i] = true;
				}

				if(!visited[i])
					unvisited++;
			}			
		} // for
		upper = min(upper, maxupper);

	} // while
	cerr << endl;
	
	
	cerr << "A diameter path: " << endl;
	for(int i=0; i<(signed)solpath.size(); i++) {
		cerr << solpath[i] << ' ';
	}
	cerr << endl;	
	
	return lower;
} // diameter
